from typing import Callable, Tuple, Optional, Union
from typing import Iterable

import numpy as np
from numpy import outer, ndarray
from scipy.special import factorial as fact
from numba import njit

from .func import isMLSWeightFunction, ConstantWeightFunction, MLSWeightFunction


def moving_least_squares(
    points: ndarray,
    values: ndarray,
    *,
    w: Optional[Union[Callable, None]] = None,
    **kwargs,
) -> Callable:
    if not isMLSWeightFunction(w):
        dim = points.shape[1]
        w = ConstantWeightFunction(dim)

    def inner(x):
        if not isinstance(x, np.ndarray):
            if isinstance(x, Iterable):
                x = np.array(x)
            else:
                raise TypeError
        w.core = x
        f = weighted_least_squares(points, values, w=w, **kwargs)
        return f(x)

    return inner


def least_squares(
    points: ndarray,
    values: ndarray,
    *,
    deg: Optional[int] = 1,
    order: Optional[int] = 2,
) -> Callable:
    """
    Given :math:`N` points located at :math:`\mathbf{x}_i` in :math:`\mathbb{R}^d`
    where :math:`i \in [1 \dots N]`. The returned fit function approximates the given
    values :math:`f_i` at :math:`\mathbf{x}_i` in the least-squares sence with the
    error functional

    .. math::
        \sum_{i} \\biggr[ || f \left( \mathbf{x}_i \\right) - f_i || \\biggr] ^2

    where :math:`f` is taken from :math:`\Pi_{m}^d`, the space of polynomials of
    total degree :math:`m` in :math:`d` spatial dimensions.

    Parameters
    ----------
    points: Iterable
        [[X11, X12, ..., X1d], ..., [Xn1, Xn2, ..., Xnd]]
    values: Iterable
        [[f11, f12, ..., f1r], ..., [fn1, fn2, ..., fnr]]
    deg: int, Optional
        The degree of the fit function. Default is 1.
    order: int, Optional.
        The order of the approximation. Default is 2.

    Returns
    -------
    Callable
        Fit function r(x) -> f(x), fdx(x), fdy(x), fdxx(x), fdyy(x), fdxy(x)
        fi([X1, X2, ..., Xd]) = [fi1, fi2,..., fir]

    Note
    ----
    The resulting fit function can have an approximation or regression behaviour,
    depending on the dataset and the degree of the polynomial.
    """
    return weighted_least_squares(points, values, deg=deg, order=order)


def weighted_least_squares(
    points: ndarray,
    values: ndarray,
    *,
    deg: Optional[int] = 1,
    order: Optional[int] = 2,
    w: Optional[Union[Callable, None]] = None,
) -> Callable:
    """
    Returns a Callable that can be used to approximate over datasets.

    Parameters
    ----------
    points: Iterable
        [[X11, X12, ..., X1d], ..., [Xn1, Xn2, ..., Xnd]]
    values: Iterable
        [[f11, f12, ..., f1r], ..., [fn1, fn2, ..., fnr]]
    deg: int, Optional
        The degree of the fit function. Default is 1.
    w: MLSWeightFunction, Optional
        A proper weight function. Default is a `ConstantWeightFunction`.
    order: int, Optional.
        The order of the approximation. Default is 2.

    Returns
    -------
    Callable
        Fit function r(x) -> f(x), fdx(x), fdy(x), fdxx(x), fdyy(x), fdxy(x)
        fi([X1, X2, ..., Xd]) = [fi1, fi2,..., fir]

    Note
    ----
    The resulting fit function can have an approximation or regression behaviour,
    depending on the dataset and the degree of the polynomial.
    """
    if not isinstance(points, ndarray):
        points = np.array(points)

    if not isinstance(values, ndarray):
        values = np.array(values)

    assert isinstance(points, np.ndarray)
    assert isinstance(values, np.ndarray)
    assert points.shape[0] == values.shape[0]

    if len(values.shape) == 1:
        values = values.reshape(len(values), 1)

    dim = points.shape[1]

    if isMLSWeightFunction(w):
        assert dim == w.dimension
    else:
        w = ConstantWeightFunction(dim)

    grad = True if order > 0 else False
    hess = True if order > 1 else False

    if hess:
        grad = True
        assert hasattr(w, "Hessian")

    if grad:
        assert hasattr(w, "gradient")

    b, bdx, bdy, bdxx, bdyy, bdxy = _get_polynomial(deg, dim)

    (invA, V, B, Adx, Ady, Adxx, Adyy, Adxy, Bdx, Bdy, Bdxx, Bdyy, Bdxy) = _mls_preproc(
        points, values, deg, w, b, grad, hess
    )

    def inner(x):
        return _mls_approx(
            invA,
            B,
            b(x),
            V,
            Adx,
            Ady,
            Adxx,
            Adyy,
            Adxy,
            Bdx,
            Bdy,
            Bdxx,
            Bdyy,
            Bdxy,
            bdx(x),
            bdy(x),
            bdxx(x),
            bdyy(x),
            bdxy(x),
            g=grad,
            H=hess,
        )

    return inner


def _get_polynomial(deg: int, dim: int):
    if deg == 1:
        if dim == 1:

            def b(x):
                return np.array([1, x])

            def bdx(x):
                return np.array([0, 1])

            def bdy(x):
                return None

            def bdxx(x):
                return np.array([0, 0])

            def bdyy(x):
                return None

            def bdxy(x):
                return None

        elif dim == 2:

            def b(x):
                return np.array([1, x[0], x[1]])

            def bdx(x):
                return np.array([0, 1, 0])

            def bdy(x):
                return np.array([0, 0, 1])

            def bdxx(x):
                return np.array([0, 0, 0])

            def bdyy(x):
                return np.array([0, 0, 0])

            def bdxy(x):
                return np.array([0, 0, 0])

        else:
            raise NotImplementedError
    elif deg == 2:
        if dim == 1:

            def b(x):
                return np.array([1, x, x**2])

            def bdx(x):
                return np.array([0, 1, 2 * x[0]])

            def bdy(x):
                return None

            def bdxx(x):
                return np.array([0, 0, 2])

            def bdyy(x):
                return None

            def bdxy(x):
                return None

        elif dim == 2:

            def b(x):
                return np.array([1, x[0], x[1], x[0] ** 2, x[1] ** 2, x[0] * x[1]])

            def bdx(x):
                return np.array([0, 1, 0, 2 * x[0], 0, x[1]])

            def bdy(x):
                return np.array([0, 0, 1, 0, 2 * x[1], x[0]])

            def bdxx(x):
                return np.array([0, 0, 0, 2, 0, 0])

            def bdyy(x):
                return np.array([0, 0, 0, 0, 2, 0])

            def bdxy(x):
                return np.array([0, 0, 0, 0, 0, 1])

        else:
            raise NotImplementedError
    else:
        raise NotImplementedError

    return b, bdx, bdy, bdxx, bdyy, bdxy


def _mls_preproc(
    points: ndarray,
    values: ndarray,
    deg: int,
    w: MLSWeightFunction,
    b: Callable,
    g: Optional[bool] = True,
    H: Optional[bool] = True,
) -> Tuple[ndarray]:
    nData = points.shape[0]
    nDim = points.shape[1]
    nRec = values.shape[1]

    # moment matrix
    k = int(fact(deg + nDim) / fact(deg) / fact(nDim))
    A = np.zeros([k, k])
    B = np.zeros([k, nData])
    Adx, Ady, Bdx, Bdy = 4 * (None,)
    Adxx, Adyy, Adxy, Bdxx, Bdyy, Bdxy = 6 * (None,)

    if g:
        Adx = np.zeros([k, k])
        Ady = np.zeros([k, k])
        Bdx = np.zeros([k, nData])
        Bdy = np.zeros([k, nData])

    if H:
        Adxx = np.zeros([k, k])
        Adyy = np.zeros([k, k])
        Adxy = np.zeros([k, k])
        Bdxx = np.zeros([k, nData])
        Bdyy = np.zeros([k, nData])
        Bdxy = np.zeros([k, nData])

    V = np.zeros([nData, nRec])
    for i in range(nData):
        xi = points[i]
        fi = values[i]
        bi = b(xi)
        wi = w.f(xi)
        Mi = outer(bi, bi)
        A += Mi * wi
        B[:, i] += bi * wi
        if g:
            gwi = w.g(xi)
            Adx += Mi * gwi[0]
            Ady += Mi * gwi[1]
            Bdx[:, i] += bi * gwi[0]
            Bdy[:, i] += bi * gwi[1]
        if H:
            Gwi = w.G(xi)
            Adxx += Mi * Gwi[0, 0]
            Adyy += Mi * Gwi[1, 1]
            Adxy += Mi * Gwi[0, 1]
            Bdxx[:, i] += bi * Gwi[0, 0]
            Bdyy[:, i] += bi * Gwi[1, 1]
            Bdxy[:, i] += bi * Gwi[0, 1]
        for r in range(nRec):
            V[i, r] = fi[r]

    invA = np.linalg.inv(A)

    return invA, V, B, Adx, Ady, Adxx, Adyy, Adxy, Bdx, Bdy, Bdxx, Bdyy, Bdxy


def _mls_approx(
    invA: ndarray,
    B: ndarray,
    b: ndarray,
    V: ndarray,
    Adx: ndarray,
    Ady: ndarray,
    Adxx: ndarray,
    Adyy: ndarray,
    Adxy: ndarray,
    Bdx: ndarray,
    Bdy: ndarray,
    Bdxx: ndarray,
    Bdyy: ndarray,
    Bdxy: ndarray,
    bdx: ndarray,
    bdy: ndarray,
    bdxx: ndarray,
    bdyy: ndarray,
    bdxy: ndarray,
    g: Optional[bool] = True,
    H: Optional[bool] = True,
) -> Tuple[ndarray, ndarray, ndarray, ndarray, ndarray, ndarray]:
    gamma = invA @ b
    f = _mls_f(gamma, B, V)

    fdx, fdy, fdxx, fdyy, fdxy = (None,) * 5

    if g:
        gammadx = invA @ (bdx - Adx @ gamma)
        gammady = invA @ (bdy - Ady @ gamma)
        fdx, fdy = _mls_g(gamma, gammadx, gammady, B, Bdx, Bdy, V)

    if H:
        fdxx, fdyy, fdxy = _mls_H(
            invA,
            bdxx,
            bdyy,
            bdxy,
            Adx,
            Ady,
            Adxx,
            Adyy,
            Adxy,
            gamma,
            gammadx,
            gammady,
            B,
            Bdx,
            Bdy,
            Bdxx,
            Bdyy,
            Bdxy,
            V,
        )

    return f, fdx, fdy, fdxx, fdyy, fdxy


@njit(nogil=True, cache=True)
def _mls_f(
    gamma: ndarray,
    B: ndarray,
    V: ndarray,
) -> Tuple[ndarray, ndarray]:
    SHP = gamma @ B
    f = SHP.T @ V
    return f


@njit(nogil=True, cache=True)
def _mls_g(
    gamma: ndarray,
    gammadx: ndarray,
    gammady: ndarray,
    B: ndarray,
    Bdx: ndarray,
    Bdy: ndarray,
    V: ndarray,
) -> Tuple[ndarray, ndarray]:
    SHPdx = gammadx @ B + gamma @ Bdx
    SHPdy = gammady @ B + gamma @ Bdy
    fdx = SHPdx.T @ V
    fdy = SHPdy.T @ V
    return fdx, fdy


@njit(nogil=True, cache=True)
def _mls_H(
    invA: ndarray,
    bdxx: ndarray,
    bdyy: ndarray,
    bdxy: ndarray,
    Adx: ndarray,
    Ady: ndarray,
    Adxx: ndarray,
    Adyy: ndarray,
    Adxy: ndarray,
    gamma: ndarray,
    gammadx: ndarray,
    gammady: ndarray,
    B: ndarray,
    Bdx: ndarray,
    Bdy: ndarray,
    Bdxx: ndarray,
    Bdyy: ndarray,
    Bdxy: ndarray,
    V: ndarray,
) -> Tuple[ndarray, ndarray, ndarray]:
    gammadxx = invA @ (bdxx - Adxx @ gamma - 2 * Adx @ gammadx)
    gammadyy = invA @ (bdyy - Adyy @ gamma - 2 * Ady @ gammady)
    gammadxy = invA @ (bdxy - Adxy @ gamma - Adx @ gammady - Ady @ gammadx)
    SHPdxx = gammadxx @ B + 2 * gammadx @ Bdx + gamma @ Bdxx
    SHPdyy = gammadyy @ B + 2 * gammady @ Bdy + gamma @ Bdyy
    SHPdxy = gammadxy @ B + gammadx @ Bdy + gammady @ Bdx + gamma @ Bdxy
    fdxx = SHPdxx.T @ V
    fdyy = SHPdyy.T @ V
    fdxy = SHPdxy.T @ V
    return fdxx, fdyy, fdxy



# from collections.abc import Iterable
from typing import Tuple, Union, Optional, Any, Iterable
from numbers import Number

import numpy as np
from numpy.linalg import norm
from numpy import ndarray

from ..function import Function


class MLSWeightFunction(Function):
    """
    Base class for weight functions for the moving least squares method.
    """

    def __init__(self, core: Union[int, Iterable[Number], ndarray]):
        if not isinstance(core, np.ndarray):
            if isinstance(core, Iterable):
                core = np.array(core)
                dim = core.shape[0]
            elif isinstance(core, int):
                dim = core
                core = np.zeros([dim])
            else:
                raise ValueError(f"Type {type(core)} is invalid for 'core'.")
        else:
            dim = core.shape[0]
        super().__init__(d=dim)
        self.core = core

    def value(self, x: Iterable[Number]) -> float:
        """
        Evaluates the function.
        """
        if not isinstance(x, np.ndarray):
            if isinstance(x, Iterable):
                x = np.array(x)
            else:
                raise ValueError(
                    f"Expected a NumPy ndarray, or an Iterable, got {type(x)}"
                )


def isMLSWeightFunction(f: Any) -> bool:
    """
    Returns `True` if the argument is a valid weight function for the
    moving least squares method.
    """
    c1 = isinstance(f, MLSWeightFunction)
    c2 = MLSWeightFunction in list(type(f).__bases__)
    return any([c1, c2])


class ConstantWeightFunction(MLSWeightFunction):
    """
    A constant weight function for the moving least squares method.
    """

    def __init__(self, dimension: int, val: Optional[Number] = 1.0):
        super().__init__(np.zeros([dimension]))
        self.val = val
        return

    def value(self, x: Iterable[Number]) -> float:
        return self.val

    def gradient(self, x: Iterable[Number]) -> ndarray:
        return np.zeros([self.dimension])

    def Hessian(self, x: Iterable[Number]) -> ndarray:
        return np.zeros([self.dimension, self.dimension])


class SingularWeightFunction(MLSWeightFunction):
    """
    A singular weight function for the moving least squares method.
    """

    def __init__(
        self, core: Union[int, Iterable[Number], ndarray], eps: Optional[float] = 1e-5
    ):
        super().__init__(core)
        self.eps = eps
        return

    def value(self, x: Iterable[Number]):
        super().value(x)
        return 1 / (norm(np.subtract(self.core, x)) ** 2 + self.eps**2)


class CubicWeightFunction(MLSWeightFunction):
    """
    A cubic weight function for the moving least squares method.

    Example
    -------
    >>> from sigmaepsilon.math.approx import CubicWeightFunction
    >>> w = CubicWeightFunction([0.0, 0.0], [0.5, 0.5])
    >>> w([0.0, 0.0])
    0.4444444444444444
    """

    def __init__(
        self,
        core: Union[int, Iterable[Number], ndarray],
        supportdomain: Iterable[Number],
    ):
        super().__init__(core)
        assert self.dimension == 2
        self.supportdomain = supportdomain
        return

    def evaluator(self, x: Iterable[Number]) -> Tuple[float, ndarray, ndarray]:
        d = np.subtract(self.core, x)
        difX = d[0]
        difY = d[1]
        dmX = self.supportdomain[0]
        dmY = self.supportdomain[1]
        rX = abs(difX) / dmX
        rY = abs(difY) / dmY

        if abs(difX) < 1e-12:
            drdX = 0
        else:
            drdX = (difX / abs(difX)) / dmX

        if abs(difY) < 1e-12:
            drdY = 0
        else:
            drdY = (difY / abs(difY)) / dmY

        if rX <= 0.5:
            wX = 2 / 3 - 4 * rX**2 + 4 * rX**3
            dwXdX = (-8 * rX + 12 * rX**2) * drdX
            dwXdXX = (-8 + 24 * rX) * drdX * drdX
        elif rX > 0.5 and rX <= 1:
            wX = 4 / 3 - 4 * rX + 4 * rX**2 - (4 / 3) * rX**3
            dwXdX = (-4 + 8 * rX - 4 * rX**2) * drdX
            dwXdXX = (8 - 8 * rX) * drdX * drdX
        else:
            wX = 0
            dwXdX = 0
            dwXdXX = 0

        if rY <= 0.5:
            wY = 2 / 3 - 4 * rY**2 + 4 * rY**3
            dwYdY = (-8 * rY + 12 * rY**2) * drdY
            dwYdYY = (-8 + 24 * rY) * drdY * drdY
        elif rY > 0.5 and rY <= 1:
            wY = 4 / 3 - 4 * rY + 4 * rY**2 - (4 / 3) * rY**3
            dwYdY = (-4 + 8 * rY - 4 * rY**2) * drdY
            dwYdYY = (8 - 8 * rY) * drdY * drdY
        else:
            wY = 0
            dwYdY = 0
            dwYdYY = 0

        val = wX * wY
        grad = np.array([wY * dwXdX, wX * dwYdY])
        Hessian = np.array([[wY * dwXdXX, dwXdX * dwYdY], [dwXdX * dwYdY, wX * dwYdYY]])

        return val, grad, Hessian

    def value(self, x: Iterable[Number]) -> float:
        super().value(x)
        res, _, _ = self.evaluator(x)
        return res

    def gradient(self, x: Iterable[Number]) -> ndarray:
        super().value(x)
        _, res, _ = self.evaluator(x)
        return res

    def Hessian(self, x: Iterable[Number]) -> ndarray:
        super().value(x)
        _, _, res = self.evaluator(x)
        return res
